require "util"
require "utils.set"
require "utils.train"

function onInit()
    global.trains_lookup = {}
    global.last_built_sign = {}
end

function onConfigurationChanged(data)
    -- remove old unused variables
    global.trains = nil
    global.slowers = nil
    global.unslowers = nil

    -- enable recipes if tech is already researched
    for k, v in pairs(game.forces) do
        if v.technologies["rail-signals"].researched then
            v.recipes["electronic-train-limit"].enabled = true
            v.recipes["electronic-train-unlimit"].enabled = true
        end
    end
    
    global.trains_lookup = global.trains_lookup or {}
    global.last_built_sign = global.last_built_sign or {}
end

script.on_init(onInit)
script.on_configuration_changed(onConfigurationChanged)

function getBoundingBox(position, radius)
    return {{position.x - radius, position.y - radius}, {position.x + radius, position.y + radius}}
end

function isTrainArriving(train)
    return train.state == defines.train_state.arrive_station or train.state == defines.train_state.arrive_signal
end

function shouldBrakeTrain(train, allowed_speed)
    -- for train braking before station or signal we're adding additional ~10 km/h to the target speed
    -- this way it will not roll slowly for a half of minute before reaching the station/signal
    return not train.manual_mode and ((not isTrainArriving(train) and math.abs(train.speed) >= allowed_speed) or (isTrainArriving(train) and math.abs(train.speed) >= allowed_speed + 0.05))
end

function calcDeceleration(train)
    local abs_speed = math.abs(train.speed)
    local factor = 1

    if isTrainArriving(train) then
        factor = 0.75
    end

    if abs_speed > 0.6 then
        return 0.005*factor
    elseif abs_speed > 0.3 then
        return 0.0075*factor
    else
        return 0.01*factor
    end
end

function calcTickSpeed(speed)
    -- convert from speed in kph to speed in tiles per tick
    return speed / 216.4
end

function processSignsForTrain(twrap)
    -- first locomotive of the train moving in proper direction
    local loco = nil

    if twrap.train.speed > 0 and #twrap.train.locomotives.front_movers > 0 then
        loco = twrap.train.locomotives.front_movers[1]
    elseif twrap.train.speed < 0 and #twrap.train.locomotives.back_movers > 0 then
        loco = twrap.train.locomotives.back_movers[#twrap.train.locomotives.back_movers]
    else
        -- invalid setup, no locomotive in the movement direction of train
        return
    end

    lst = loco.surface.find_entities_filtered{area = getBoundingBox(loco.position, 2), type = "constant-combinator"}
    for i, ent in pairs(lst) do
        if ent.name == "placed-train-limit" and isOnLeft(ent.position, loco.position, getCardinal(loco.orientation)) then
            twrap.limit = calcTickSpeed(getLimitIndication(ent))
            return
        elseif ent.name == "placed-train-unlimit" and isOnLeft(ent.position, loco.position, getCardinal(loco.orientation)) then
            twrap.limit = false
            return
        end
    end

    local area = getBoundingBox(loco.position, twrap.train.speed * 10 + 1)

    if loco.surface.count_entities_filtered{area = area, type = "constant-combinator", count = 1} == 0 then
        twrap.deactivation = 10
    end
end

script.on_event(defines.events.on_tick, function(event)
    for i=#global.trains_lookup,1,-1 do
        local twrap = global.trains_lookup[i]

        if not twrap.train.valid then
            table.remove(global.trains_lookup, i)
        else
            -- process signs and deactivation for this train
            if twrap.deactivation == 0 and twrap.train.speed ~= 0 then
                processSignsForTrain(twrap)
            elseif twrap.deactivation > 0 then
                 twrap.deactivation = twrap.deactivation - 1
            elseif twrap.deactivation < 0 then
                twrap.deactivation = 0
            end

            -- decelerate the train if needed
            if twrap.limit and twrap.train.speed ~= 0 and shouldBrakeTrain(twrap.train, twrap.limit) then
                local decel = calcDeceleration(twrap.train)
                if twrap.train.speed > 0 then
                    twrap.train.speed = math.max(twrap.train.speed - decel, twrap.limit)
                else
                    twrap.train.speed = math.min(twrap.train.speed + decel, twrap.limit)
                end
            end
        end
    end
end)

function onTrainState(event)
    for k, v in pairs(global.trains_lookup) do
        if v.train == event['train'] then
            return
        end
    end

    table.insert(global.trains_lookup, {train=event['train'], deactivation=0, limit=false})
end

function getLimitIndication(entity)
    return entity.get_or_create_control_behavior().parameters.parameters[1].count
end

function setLimitIndication(entity, value)
    if entity and entity.valid then
        entity.get_or_create_control_behavior().parameters = {parameters={
            {index = 1, count = value, signal = {type="item", name="train-limit"}}
        }}
    end
end

function onBuiltEntity(event)
    local player = nil
    
    if event.player_index then
        player = game.players[event.player_index]
    end

    if event.created_entity.name == "placed-train-limit" then
        event.created_entity.operable = false
        
        if player then
            setLimitIndication(event.created_entity, 20)

            local frame = player.gui.center["tsl-value"]

            if not frame then
                global.last_built_sign[event.player_index] = event.created_entity

                frame = player.gui.center.add{type="frame", name="tsl-value", caption={"gui-limit-title"}, direction="horizontal"}
                frame.add{type="button", name="tsl-value-10", caption={"gui-10"}}
                frame.add{type="button", name="tsl-value-20", caption={"gui-20"}}
                frame.add{type="button", name="tsl-value-50", caption={"gui-50"}}
                frame.add{type="button", name="tsl-value-80", caption={"gui-80"}}
                frame.add{type="button", name="tsl-value-100", caption={"gui-100"}}
                frame.add{type="button", name="tsl-value-other", caption={"gui-other"}}
            end
        end
    elseif event.created_entity.name == "placed-train-unlimit" then
        event.created_entity.operable = false
    end
end

function onGuiClick(event)
    local player = game.players[event.player_index]
	local frame = player.gui.center["tsl-value"]
    local manual_frame = player.gui.center["tsl-manual-value"]
    
    if manual_frame and event.element.parent == manual_frame then
        if event.element.name == "tsl-manual-ok" then
            local limit_value = tonumber(manual_frame["tsl-manual-input"].text)
            
            if not limit_value or limit_value < 5 or limit_value > 500 or limit_value % 1 ~= 0 then
                manual_frame["tsl-manual-input"].text = ""
                player.print({"gui-invalid-limit"})
                return
            end

            setLimitIndication(global.last_built_sign[event.player_index], limit_value)
            global.last_built_sign[event.player_index] = nil
            
            manual_frame.destroy()
        end
    elseif frame and event.element.parent == frame then
        local limit_value = nil

        if event.element.name == "tsl-value-other" then
            frame.destroy()

            manual_frame = player.gui.center.add{type="frame", name="tsl-manual-value", caption={"gui-limit-manual"}, direction="vertical"}
            manual_frame.add{type="label", caption={"gui-limit-value"}}
            manual_frame.add{type="textfield", name="tsl-manual-input"}
            manual_frame.add{type="button", name="tsl-manual-ok", caption={"gui-ok"}}
            return
        elseif event.element.name == "tsl-value-10" then
            limit_value = 10
        elseif event.element.name == "tsl-value-20" then
            limit_value = 20
        elseif event.element.name == "tsl-value-50" then
            limit_value = 50
        elseif event.element.name == "tsl-value-80" then
            limit_value = 80
        elseif event.element.name == "tsl-value-100" then
            limit_value = 100
        else
            return
        end

        setLimitIndication(global.last_built_sign[event.player_index], limit_value)
        global.last_built_sign[event.player_index] = nil
        
        frame.destroy()
    end
end

function onPlayerLeft(event)
    global.last_built_sign[event.player_index] = nil
end

script.on_event(defines.events.on_train_changed_state, onTrainState)
script.on_event(defines.events.on_built_entity, onBuiltEntity)
script.on_event(defines.events.on_robot_built_entity, onBuiltEntity)
script.on_event(defines.events.on_gui_click, onGuiClick)
script.on_event(defines.events.on_player_left_game, onPlayerLeft)
